"""
Brendan Croom
brendan.croom.ctr@us.af.mil

TODO:
+ Implement subpixel search - Gaussian fitting?

+++++++++++++++++++++++++

Overview of code:

Standard analysis:
+ piv_folder: Performs PIV calculation on a sequence of images in a folder
  + piv_image_pair: Performs PIV on a pair of images... called by "piv_folder"
    + piv_search: Performs PIV search for a single template... called by "piv_image_pair"

Alternately:
+ piv_folder_initial_guess: Similar to "piv_folder", but with precise initial guesses

"""

import os, gc, time, sys, warnings
import imageio as io
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import matplotlib.tri as tri
import scipy.signal, scipy.interpolate, scipy.optimize
from joblib import Parallel, delayed
from skimage.filters import median


# Suppress warnings:
if not sys.warnoptions:
    warnings.simplefilter("ignore")


def average_images_in_folder(folder, batch_size=25):
    """
    A scalable way to average many images into one

    Edited: 16 JAN 2020
    """
    def average_images(list_of_images):
        return np.average(list_of_images, axis=0)
    def average_file_list(list_of_files):
        img_list = [io.imread(f).astype('uint16') for f in list_of_files]
        return average_images(img_list)

    # Build a list of files to process:
    file_list = [os.path.join(folder, f) for f in os.listdir(folder) if f.endswith('.tiff')]

    result = None
    step_list = range(0, len(file_list), batch_size)
    for i in range(0, len(file_list), batch_size):
        img_i = average_file_list(file_list[slice(i, i+batch_size)]).astype(float)

        if result is None:
            result = img_i / len(step_list)
        else:
            result = result + img_i / len(step_list)

    return result


# Define several useful correlation algorithms:
def corr_leastsq(template, img):
    raise NotImplementedError


def corr_xcorr(template, img):
    """
    Computes the cross-correlation between "template" and "img"

    Edited: 16 JAN 2020
    """
    corr_result = scipy.signal.correlate2d(template.astype(float), img.astype(float), mode='valid')
    return corr_result
        

def corr_xcorr_norm(template, img, FFT=False):
    """
    Computes the normalized cross-correlation between "template" and "img"

    Edited: 16 JAN 2020
    """
    template_norm = (template.astype(float) - template.mean()) / template.std()
    img_norm = (img.astype(float) - img.mean()) / img.std()
    
    corr_result = scipy.signal.correlate2d(
        template_norm, img_norm, mode='valid')
    return corr_result


def disp_from_corr_result(corr_result, interp=True):
    """
    Calculates (u, v) displacement from the correlation peak. Window_size is a
    tuple, showing dimensions of the search template.

    Optional: perform subpixel interpolation

    Note:
    + Assumes that the facet is aligned with the center of the search window
    + Assumes size of the window = 2M+1, where M = window_halfsize
    + Assumes size of the search area = 2N+1

    edited: 16 JAN 2020
    """
    # Find the peak location to nearest integer value:
    peak_location = np.array(np.unravel_index(np.argmax(corr_result), corr_result.shape))

    # Interpolate around this region (optional):
    if interp: 
        # Build a small interpolation function to minimize
        peak_cropped = crop_img(corr_result, peak_location, 7)  # May need to check edge cases...
        x, y = peak_location.reshape((2, -1)) + np.tile(np.arange(-3, 4), (2, 1))
        peak_fit = scipy.interpolate.interp2d(x, y, peak_cropped, kind='cubic')

        # Find the peak:
        f_obj = lambda xy : -peak_fit(*xy)
        peak_location = scipy.optimize.minimize(f_obj, peak_location).x


    # Convert to (u, v) displacement:
    corr_halfsize = (np.array(corr_result.shape) - 1) / 2
    disp = peak_location - corr_halfsize
    return disp


def build_grid_points(img, spacing, search_halfsize):
    """
    Build a grid of search points. The grid is defined by the image size.

    edited: 20 DEC 2019
    """
    lower_lims = np.ones(2) * search_halfsize
    upper_lims = np.array(img.shape) - search_halfsize
    grid_slices = [slice(lower, upper, spacing) \
                   for lower, upper in zip(lower_lims, upper_lims)]
    x, y = np.mgrid[tuple(grid_slices)]
    
    return np.vstack((x.flatten(), y.flatten())).T


def build_grid_points_from_mask(img, spacing, search_halfsize):
    """
    Build a grid of search points - but keep only where "img > 0"

    edited: 16 JAN 2020
    """
    # Initialize a list of points, then filter out where img == 0
    xy_search = build_grid_points(img, spacing, search_halfsize)
    idx_filter = img[tuple(xy_search.T.astype(int))] > 0
    return xy_search[idx_filter]

 

def crop_img(img, xy, size):
    """
    Crop the image centered at (x, y) into a square of dimension "size"

    edited: 20 DEC 2019
    """
    assert size % 2  # check if odd
    half_size = int((size - 1 ) / 2)
    crop = [slice(coor - half_size, coor + half_size + 1) for coor in xy.astype(int)]
    return img[tuple(crop)]


def plot_PIV_results(xy_search, displacements):
    """
    Plot the PIV results using Matplotlib
    
    edited: 16 JAN 2020
    """
    x, y = xy_search.T
    u, v = displacements.T
    
    f, ax = plt.subplots(figsize=(6,6))
    if False:  # Use a tricontourf
        triangles = tri.Triangulation(*xy_search[..., ::-1].T)
        mask = np.any(np.where(np.isnan(u)[triangles.triangles], True, False), axis=1)
        triangles.set_mask(mask)
        contours = ax.tricontourf(triangles, u, 50)
    else:
        contours = ax.scatter(y, x, c=u, lw=0)
    
    ax.invert_yaxis()
    cb = f.colorbar(contours, ax=ax)
    cb.set_label('Velocity (px / img)', rotation=270, va='bottom')
    ax.set_xlabel('X (px)')
    ax.set_ylabel('Z (px)')
    ax.set_aspect('equal')
    f.tight_layout()
    return f, ax


def save_PIV_results(outpath, data):
    """
    "Intelligently" save the data as a CSV to outpath

    edited: 16 JAN 2020
    """
    try:
        np.savetxt(outpath, data)
    except FileNotFoundError:
        outdir, file = os.path.split(outpath)
        os.makedirs(outdir)
        np.savetxt(outpath, data)
    except FileExistsError:
        raise
    


def piv_search(template, search_region, method='norm_xcorr'):
    """
    Performs a PIV search of a single template within a single search region.

    edited: 20 DEC 2019
    """

    # Use the default method:
    if method is None:    
        method == 'norm_xcorr'
    if method == 'norm_xcorr':
        correlation = corr_xcorr_norm(template, search_region)
    else:
        raise NotImplementedError('Need to implement other correlation techniques')

    # Calculate the displacement:
    return disp_from_corr_result(correlation)


def piv_image_pair(img_a, img_b, xy_search, window_size, search_size, initial_guess=np.array([0, 0])):
    """
    Performs PIV velocimetry analysis between img_a to img_b.

    Parameters:
    xy_search: (N, 2) list of coordinates to analyze
    window_size = side dimension of the square search template
    search_size = side dimension of the square search region.

    edited: 20 DEC 2019
    """

    assert search_size > window_size
    from itertools import count

    displacements = np.zeros(xy_search.shape)

    current_guess = np.copy(initial_guess)
    row_start = True
    if initial_guess.size == xy_search.size:  # then we have a full list of initial guesses!
        for i, xy, current_guess in zip(count(), xy_search, initial_guess):
            template = crop_img(img_a, xy, window_size)
            search_region = crop_img(img_b, xy - current_guess, search_size)

            # Compute the displacements:
            try:
                displacements[i] = piv_search(template, search_region) + current_guess
                current_guess = np.copy(displacements[i]).astype(int)
            except ValueError:
                displacements[i] = np.array([np.nan, np.nan])
    ##            print('Error in Point {}: {}'.format(i, xy))

    else:  # Then we have a single initial guess
        for i, xy in enumerate(xy_search):
            if row_start:
                current_guess = np.copy(initial_guess)
            template = crop_img(img_a, xy, window_size)
            search_region = crop_img(img_b, xy - current_guess, search_size)

            # Compute the displacements:
            try:
                displacements[i] = piv_search(template, search_region) + current_guess
                current_guess = np.copy(displacements[i]).astype(int)
            except ValueError:
                displacements[i] = np.array([np.nan, np.nan])
    ##            print('Error in Point {}: {}'.format(i, xy))

            if row_start:
                row_start = False
            elif i+1 < len(xy_search) and xy_search[i+1, 1] < xy[1]:
                row_start = True
        
    return displacements


def piv_folder(folder, outfolder, file_template, i, window_halfsize, search_halfsize,
                 img_step=1, initial_guess=np.array([0, 0]), avg_img=None):
    """
    Correlate a pair of images located in "folder" in a way that's amenable to
    massively-parallel computation.

    Inputs:
    + folder: where the images are located
    + outfolder: where to save the PIV results
    + file_template: string that is used to identify folder.
    + i: ID of the "reference" image, i.e., file_template.format(i)
    + img_step: "deformed" image = i + img_step
    + window_halfsize, search_halfsize: Dimensions of the facet, and the search region
    + initial_guess: An estimate to narrow down the search region

    Edited: 4 JUN 2020
    """
##    file_template = 'image_{:04d}.tiff' # format of images saved in "folder"
    
    spacing = window_halfsize # spacing between correlation facets

    # Load the images:
    try:
        frame_a = io.imread(os.path.join(folder, file_template.format(i))).astype(float)
        frame_b = io.imread(os.path.join(folder, file_template.format(i+img_step))).astype(float)
    except FileNotFoundError:
        print('FileNotFound: {}'.format(os.path.join(folder, file_template.format(i))))
        return -1

    # Optional: background-correct the image using avg_img:
    if avg_img is not None:
        frame_a = frame_a - avg_img
        frame_b = frame_b - avg_img

    # Build a grid of search locations from a mask:
    img_mask = io.imread(os.path.join(os.path.split(folder)[0], 'mask.png'))[..., 0]
    xy_search = build_grid_points_from_mask(img_mask, spacing, search_halfsize)

    # perform PIV:
    print('Correlating file {}'.format(file_template.format(i)))
    t1 = time.time()
    displacements = piv_image_pair(
        frame_a, frame_b,
        xy_search,
        2*window_halfsize+1, 2*search_halfsize+1,
        initial_guess=initial_guess
    )
    t2 = time.time()
    print('PIV complete: {} points, t = {:02.2f} seconds.'.format(len(xy_search), t2-t1))
    print('{:02.3f} seconds per point'.format((t2-t1)/len(xy_search)))

    # Save the displacement data:
    result_arr = np.hstack((xy_search, displacements))
    save_PIV_results(
        os.path.join(outfolder, 'piv_{}_step{}.csv'.format(file_template.format(i)[:-5], img_step)),
        result_arr
        )

    # Plot the results:
    f, ax = plot_PIV_results(xy_search, displacements / img_step)
    ax.set_title(file_template.format(i))
    f.savefig(
        os.path.join(outfolder, 'piv_{}_step{}.png'.format(file_template.format(i)[:-5], img_step)),
        dpi=240
        )
    plt.close(f)
    
    gc.collect()
    return t2 - t1


def piv_folder_initial_guess(folder, outfolder, file_template, i, window_halfsize, search_halfsize,
                             xy_search, uv_guess, img_step=1, avg_img=None):
    """
    Correlate two consecutive images in a way that's amenable to
    massively-parallel computation. Compared to "piv_parallel", this
    method allows you to input an initial guess for each search location.

    Inputs:
    + folder: where the images are located
    + outfolder: where to save the PIV results
    + file_template: string that is used to identify folder.
    + i: ID of the "reference" image, i.e., file_template.format(i)
    + window_halfsize, search_halfsize: Dimensions of the facet, and the search region
    + xy_search: Coordinates for the search window (reference image)
    + uv_guess: Guess displacement for the motion of each search window
    + img_step: "deformed" image = i + img_step
    
    Edited: 4 JUN 2020
    """
##    file_template = 'image_{:04d}.tiff' # format of images saved in "folder"
    
    # Load the images:
    try:
        frame_a = io.imread(os.path.join(folder, file_template.format(i))).astype(float)
        frame_b = io.imread(os.path.join(folder, file_template.format(i+img_step))).astype(float)
    except FileNotFoundError:
        print('FileNotFound: {}'.format(os.path.join(folder, file_template.format(i))))
        return -1

    # Optional: background-correct the image using avg_img:
    if avg_img is not None:
        frame_a = frame_a - avg_img
        frame_b = frame_b - avg_img

    # perform PIV:
    print('Correlating file {}'.format(file_template.format(i)))
    t1 = time.time()
    displacements = piv_image_pair(
        frame_a, frame_b,
        xy_search,
        2*window_halfsize+1, 2*search_halfsize+1,
        initial_guess=uv_guess
    )
    t2 = time.time()
    print('PIV complete: {} points, t = {:02.2f} seconds.'.format(len(xy_search), t2-t1))
    print('{:02.3f} seconds per point'.format((t2-t1)/len(xy_search)))

    # Save the displacement data:
    result_arr = np.hstack((xy_search, displacements))
    save_PIV_results(
        os.path.join(outfolder, 'refined_piv_{}_step{}.csv'.format(file_template.format(i)[:-5], img_step)),
        result_arr
        )

    # Plot the results:
    f, ax = plot_PIV_results(xy_search, displacements / img_step)
    ax.set_title(file_template.format(i))
    f.savefig(
        os.path.join(outfolder, 'refined_piv_{}_step{}.png'.format(file_template.format(i)[:-5], img_step)),
        dpi=240
        )
    plt.close(f)
    
    gc.collect()
    return t2 - t1


def average_PIV_results(data_folder, img_step, prefix=None, median=False):
    """
    Read all results located in "data_folder", average, and then save the results
    into the parent directory.

    Edited: 16 JAN 2019
    """

    # Load all of the results:
    result_files = [f for f in os.listdir(data_folder) if f.endswith('.csv')]
    if prefix:
        result_files = [f for f in result_files if f.startswith(prefix)]
        
    
    df = pd.DataFrame(
        data = np.vstack(
            [np.loadtxt(os.path.join(data_folder, f)) for f in result_files]
            ),
        columns=['x', 'y', 'u', 'v']
        )

    # Find the unique (x, y) coordinates
    unique_coors = set([(x, y) for x, y in zip(df.x, df.y)])

    # Create a new dataframe with point-wise averaged data:
    unique_results = np.zeros((len(unique_coors), 4))
    for i, (x, y) in enumerate(unique_coors):
        unique_results[i, :2] = x, y
        idx_filter = (df.x == x) & (df.y == y) & (~np.isnan(df.u)) & (~np.isnan(df.v))
        unique_results[i, 2:] = df.loc[idx_filter, ('u', 'v')].values.mean(axis=0)

    # Save the averaged data:
    output_folder, subfolder = os.path.split(data_folder)
    if prefix:
        np.savetxt(os.path.join(output_folder, prefix + '_' + 'piv_results_avg_{}.csv'.format(subfolder)),
                   unique_results,
                   header='x,y,u,v', delimiter=',', comments=''
                   )
    else:
        np.savetxt(os.path.join(output_folder, 'piv_results_avg_{}.csv'.format(subfolder)),
                   unique_results,
                   header='x,y,u,v', delimiter=',', comments=''
                   )

    # And plot the results:
    xy_search = unique_results[:, :2]
    displacements = unique_results[:, 2:]
    f, ax = plot_PIV_results(xy_search, displacements / img_step)
    ax.set_title(subfolder)
    if prefix:
        f.savefig(
            os.path.join(output_folder, prefix + '_' + 'piv_averaged_{}.png'.format(subfolder)),
            dpi=600
            )
    else:
        f.savefig(
            os.path.join(output_folder, 'piv_averaged_{}.png'.format(subfolder)),
            dpi=600
            )
    plt.close(f)


def initialize_grid_from_coors(x, y):
    """
    Read (X, Y) coordinate data, and initializes a blank 2D array to represent the data.
    Assumes uniform grid pitch.
    """
    x_u, y_u = np.unique(x), np.unique(y)
    return np.zeros((x_u.size, y_u.size))

    
def map_arr_to_img(x, y):
    """
    Create an index mapping such that a vector xy = (N, 2) can populate an img = (m, n)
    """
    x_u, y_u = np.unique(x), np.unique(y)
    idx_list = np.zeros((x.size, 2)).astype(int)
    for i, x_i in enumerate(x_u):
        x_idx = np.argwhere(x == x_i).reshape(-1)
        for j, y_i in enumerate(y_u):
            y_idx = np.argwhere(y[x_idx] == y_i)
            idx_list[x_idx[y_idx]] = [i, j]
    return idx_list


def refine_piv_results(piv_data):
    """
    Read the "piv_results_avg_*.csv" file,
    Then improve the data quality via median filter

    Assumes piv_data is a pandas.DataFrame, with keys "x, y, u, v"
    """
    piv_data.loc[np.logical_or(np.isnan(piv_data.u), piv_data.u < -1E3), 'u'] = 0
    idx_map = map_arr_to_img(piv_data.x, piv_data.y)
    u_img = initialize_grid_from_coors(piv_data.x, piv_data.y).astype(int)
    u_img[tuple(idx_map.T)] = piv_data.u.round(0)

    # Apply a median filter using a cross structuring element:
    selem_cross = np.zeros((3,3)).astype(int)
    selem_cross[1] = 1; selem_cross[:, 1] = 1

    if piv_data.u.mean() < 0:
##        u_img = np.clip(-u_img, 0, np.inf)
        u_med = median(-u_img, selem_cross)
    else:
##        u_img = np.clip(u_img, 0, np.inf)
        u_med = median(u_img, selem_cross)


    # Update the database:
    piv_data.u = u_img[tuple(idx_map.T)]
    piv_data.v = piv_data.v.round(0)
    return piv_data
